package api

import (
	"context"
	"errors"
	"time"

	ppb "code.fbi.h-da.de/cocsn/api/go/gosdn/pnd"
	spb "code.fbi.h-da.de/cocsn/api/go/gosdn/southbound"
	tpb "code.fbi.h-da.de/cocsn/api/go/gosdn/transport"
	nbi "code.fbi.h-da.de/cocsn/gosdn/northbound/client"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/viper"

	pb "code.fbi.h-da.de/cocsn/api/go/gosdn/core"
	"google.golang.org/grpc"
)

var dialOptions []grpc.DialOption

func init() {
	dialOptions = []grpc.DialOption{
		grpc.WithInsecure(),
	}
}

// Init initialises the CLI client.
func Init(addr string) error {
	ctx := context.Background()
	resp, err := getAllCore(ctx, addr)
	if err != nil {
		return err
	}
	if len(resp.Pnd) > 0 {
		pid := resp.Pnd[0].Id
		viper.Set("CLI_PND", pid)
		log.Infof("PND: %v", pid)
		if len(resp.Pnd[0].Sbi) != 0 {
			sbi := resp.Pnd[0].Sbi[0].Id
			viper.Set("CLI_SBI", sbi)
			log.Infof("SBI: %v", sbi)
		}
	}
	return viper.WriteConfig()
}

// GetIds requests all UUID information from the controller
func GetIds(addr string) ([]*ppb.PrincipalNetworkDomain, error) {
	ctx := context.Background()
	resp, err := getAllCore(ctx, addr)
	if err != nil {
		return nil, err
	}
	return resp.Pnd, nil
}

func getAllCore(ctx context.Context, addr string) (*pb.GetResponse, error) {
	coreClient, err := nbi.CoreClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	req := &pb.GetRequest{
		Timestamp: time.Now().UnixNano(),
		All:       true,
	}
	return coreClient.Get(ctx, req)
}

// AddPnd takes a name, description and SBI UUID to create a new
// PrincipalNetworkDomain on the controller
func AddPnd(addr, name, description, sbi string) (*pb.SetResponse, error) {
	coreClient, err := nbi.CoreClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	ctx := context.Background()
	req := &pb.SetRequest{
		Timestamp: time.Now().UnixNano(),
		Pnd: []*pb.SetPnd{
			{
				Name:        name,
				Description: description,
				Sbi:         sbi,
			},
		},
	}

	return coreClient.Set(ctx, req)
}

// GetPnd requests one or several PrincipalNetworkDomains from the
// controller. To request all PrincipalNetworkDomains without providing
// names or UUIDs use GetIds()
func GetPnd(addr string, args ...string) (*pb.GetResponse, error) {
	coreClient, err := nbi.CoreClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	if len(args) <= 0 {
		return nil, errors.New("not enough arguments")
	}
	ctx := context.Background()
	req := &pb.GetRequest{
		Timestamp: time.Now().UnixNano(),
		Pid:       args,
	}
	return coreClient.Get(ctx, req)
}

// GetChanges requests all pending and unconfirmed changes from the controller
func GetChanges(addr, pnd string) (*ppb.GetResponse, error) {
	ctx := context.Background()
	client, err := nbi.PndClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	req := &ppb.GetRequest{
		Timestamp: time.Now().UnixNano(),
		Request: &ppb.GetRequest_Change{
			Change: &ppb.GetChange{
				All: true,
			},
		},
		Pid: pnd,
	}
	return client.Get(ctx, req)
}

// Commit sends a commit request for one or multiple changes to the
// controller.
func Commit(addr, pnd string, cuids ...string) (*ppb.SetResponse, error) {
	changes := make([]*ppb.SetChange, len(cuids))
	for i, arg := range cuids {
		changes[i] = &ppb.SetChange{
			Cuid: arg,
			Op:   ppb.SetChange_COMMIT,
		}
	}
	return commitConfirm(addr, pnd, changes)
}

// Confirm sends a confirm request for one or multiple changes to the
// controller
func Confirm(addr, pnd string, cuids ...string) (*ppb.SetResponse, error) {
	changes := make([]*ppb.SetChange, len(cuids))
	for i, arg := range cuids {
		changes[i] = &ppb.SetChange{
			Cuid: arg,
			Op:   ppb.SetChange_CONFIRM,
		}
	}
	return commitConfirm(addr, pnd, changes)
}

func commitConfirm(addr, pnd string, changes []*ppb.SetChange) (*ppb.SetResponse, error) {
	ctx := context.Background()
	client, err := nbi.PndClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	req := &ppb.SetRequest{
		Timestamp: time.Now().UnixNano(),
		Change:    changes,
		Pid:       pnd,
	}
	return client.Set(ctx, req)
}

// AddDevice adds a new device to the controller. The device name is optional.
// If no name is provided a name will be generated upon device creation.
func AddDevice(addr, username, password, sbi, pnd, deviceAddress, deviceName string) (*ppb.SetResponse, error) {
	pndClient, err := nbi.PndClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}

	req := &ppb.SetRequest{
		Timestamp: time.Now().UnixNano(),
		Ond: []*ppb.SetOnd{
			{
				Address: deviceAddress,
				Sbi: &spb.SouthboundInterface{
					Id: sbi,
				},
				DeviceName: deviceName,
				TransportOption: &tpb.TransportOption{
					Address:  addr,
					Username: username,
					Password: password,
					TransportOption: &tpb.TransportOption_GnmiTransportOption{
						GnmiTransportOption: &tpb.GnmiTransportOption{},
					},
				},
			},
		},
		Pid: pnd,
	}
	ctx := context.Background()
	return pndClient.Set(ctx, req)
}

// GetDevice requests one or multiple devices belonging to a given
// PrincipalNetworkDomain from the controller. If no device identifier
// is provided, all devices are requested.
func GetDevice(addr, pid, path string, did ...string) (*ppb.GetResponse, error) {
	pndClient, err := nbi.PndClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}

	var all bool
	if len(did) == 0 {
		all = true
	}

	req := &ppb.GetRequest{
		Timestamp: time.Now().UnixNano(),
		Request: &ppb.GetRequest_Ond{
			Ond: &ppb.GetOnd{
				All: all,
				Did: did,
			},
		},
		Pid: pid,
	}
	ctx := context.Background()
	return pndClient.Get(ctx, req)
}

// Update creates a ChangeRequest to update the given path with the given value
// at the given OND on the controller.
func Update(addr, did, pid, path, value string) (*ppb.SetResponse, error) {
	req := &ppb.ChangeRequest{
		Id:    did,
		Path:  path,
		Value: value,
		ApiOp: ppb.ApiOperation_UPDATE,
	}
	return sendChangeRequest(addr, pid, req)
}

// Replace creates a ChangeRequest to replace the given path with the given value
// at the given OND on the controller.
func Replace(addr, did, pid, path, value string) (*ppb.SetResponse, error) {
	req := &ppb.ChangeRequest{
		Id:    did,
		Path:  path,
		Value: value,
		ApiOp: ppb.ApiOperation_REPLACE,
	}
	return sendChangeRequest(addr, pid, req)
}

// Delete creates a ChangeRequest to delete the given path node
// at the given OND on the controller.
func Delete(addr, did, pid, path string) (*ppb.SetResponse, error) {
	req := &ppb.ChangeRequest{
		Id:    did,
		Path:  path,
		ApiOp: ppb.ApiOperation_DELETE,
	}
	return sendChangeRequest(addr, pid, req)
}

func sendChangeRequest(addr, pid string, req *ppb.ChangeRequest) (*ppb.SetResponse, error) {
	pndClient, err := nbi.PndClient(addr, dialOptions...)
	if err != nil {
		return nil, err
	}
	ctx := context.Background()
	r := &ppb.SetRequest{
		Timestamp:     time.Now().UnixNano(),
		ChangeRequest: []*ppb.ChangeRequest{req},
		Pid:           pid,
	}
	return pndClient.Set(ctx, r)
}
