package server

import (
	"context"
	"reflect"
	"time"

	"code.fbi.h-da.de/cocsn/gosdn/interfaces/device"

	ppb "code.fbi.h-da.de/cocsn/api/go/gosdn/pnd"
	spb "code.fbi.h-da.de/cocsn/api/go/gosdn/southbound"
	"code.fbi.h-da.de/cocsn/gosdn/interfaces/networkdomain"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus/errors"
	"github.com/google/uuid"
	"github.com/openconfig/ygot/ygot"
)

type pndServer struct {
	ppb.UnimplementedPndServer
}

func (p pndServer) Get(ctx context.Context, request *ppb.GetRequest) (*ppb.GetResponse, error) {
	pid, err := uuid.Parse(request.Pid)
	if err != nil {
		return nil, err
	}
	switch req := request.Request.(type) {
	case *ppb.GetRequest_Pnd:
		return handleGetPnd(pid)
	case *ppb.GetRequest_Ond:
		return handleGetOnd(pid, req)
	case *ppb.GetRequest_Sbi:
		return handleGetSbi(pid, req)
	case *ppb.GetRequest_Change:
		return handleGetChange(pid, req)
	default:
		return nil, errors.ErrOperationNotSupported{Op: reflect.TypeOf(request.Request)}
	}
}

func handleGetPnd(pid uuid.UUID) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, err
	}
	onds, err := fillOnds(pnd, true)
	if err != nil {
		return nil, err
	}
	sbis, err := fillSbis(pnd, true)
	if err != nil {
		return nil, err
	}
	changes, err := fillChanges(pnd, true)
	if err != nil {
		return nil, err
	}

	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Pnd: &ppb.PrincipalNetworkDomain{
			Id:          pnd.ID().String(),
			Name:        pnd.GetName(),
			Description: pnd.GetDescription(),
			Ond:         onds,
			Sbi:         sbis,
			Change:      changes,
		},
	}, nil
}

func handleGetSbi(pid uuid.UUID, req *ppb.GetRequest_Sbi) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, err
	}
	sbis, err := fillSbis(pnd, req.Sbi.GetAll(), req.Sbi.Sid...)
	if err != nil {
		return nil, err
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Sbi:       sbis,
	}, nil
}

func stringToUUID(sid []string) ([]uuid.UUID, error) {
	UUIDs := make([]uuid.UUID, len(sid))
	for i, id := range sid {
		parsed, err := uuid.Parse(id)
		if err != nil {
			return nil, err
		}
		UUIDs[i] = parsed
	}
	return UUIDs, nil
}

func handleGetOnd(pid uuid.UUID, req *ppb.GetRequest_Ond) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, err
	}
	onds, err := fillOnds(pnd, req.Ond.All, req.Ond.Did...)
	if err != nil {
		return nil, err
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Ond:       onds,
	}, nil
}

func handleGetChange(pid uuid.UUID, req *ppb.GetRequest_Change) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, err
	}
	changes, err := fillChanges(pnd, req.Change.All, req.Change.Cuid...)
	if err != nil {
		return nil, err
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Change:    changes,
	}, nil
}

func fillSbis(pnd networkdomain.NetworkDomain, all bool, sid ...string) ([]*spb.SouthboundInterface, error) {
	var sbiList []uuid.UUID

	sbiStore := pnd.GetSBIs().(*nucleus.SbiStore)
	switch all {
	case true:
		sbiList = sbiStore.UUIDs()
	default:
		var err error
		if len(sid) == 0 {
			return nil, &errors.ErrInvalidParameters{
				Func:  fillSbis,
				Param: "length of 'sid' cannot be '0' when 'all' is set to 'false'",
			}
		}
		sbiList, err = stringToUUID(sid)
		if err != nil {
			return nil, err
		}
	}
	sbis := make([]*spb.SouthboundInterface, len(sbiList))
	for i, id := range sbiList {
		sbi, err := sbiStore.GetSBI(id)
		if err != nil {
			return nil, err
		}
		sbis[i] = &spb.SouthboundInterface{
			Id:   id.String(),
			Type: sbi.Type(),
		}
	}
	return sbis, nil
}

func fillOnds(pnd networkdomain.NetworkDomain, all bool, did ...string) ([]*ppb.OrchestratedNetworkingDevice, error) {
	var ondList []uuid.UUID

	switch all {
	case true:
		ondList = pnd.Devices()
	default:
		var err error
		if len(did) == 0 {
			return nil, &errors.ErrInvalidParameters{
				Func:  fillOnds,
				Param: "length of 'did' cannot be '0' when 'all' is set to 'false'",
			}
		}
		ondList, err = stringToUUID(did)
		if err != nil {
			return nil, err
		}
	}

	onds := make([]*ppb.OrchestratedNetworkingDevice, len(ondList))
	for i, id := range ondList {
		d, err := pnd.GetDevice(id.String())
		if err != nil {
			return nil, err
		}
		cfg := ygot.GNMINotificationsConfig{}
		dev, err := ygot.TogNMINotifications(d.(device.Device).Model(), time.Now().UnixNano(), cfg)
		if err != nil {
			return nil, err
		}
		onds[i] = &ppb.OrchestratedNetworkingDevice{
			Id:     id.String(),
			Name:   "",
			Device: dev,
		}
	}
	return onds, nil
}

func fillChanges(pnd networkdomain.NetworkDomain, all bool, cuid ...string) ([]*ppb.Change, error) {
	var changeList []uuid.UUID

	switch all {
	case true:
		changeList = pnd.PendingChanges()
		changeList = append(changeList, pnd.CommittedChanges()...)
	default:
		var err error
		if len(cuid) == 0 {
			return nil, &errors.ErrInvalidParameters{
				Func:  fillOnds,
				Param: "length of 'did' cannot be '0' when 'all' is set to 'false'",
			}
		}
		changeList, err = stringToUUID(cuid)
		if err != nil {
			return nil, err
		}
	}

	changes := make([]*ppb.Change, len(changeList))
	for i, ch := range changeList {
		c, err := pnd.GetChange(ch)
		if err != nil {
			return nil, err
		}
		change, ok := c.(*nucleus.Change)
		if !ok {
			return nil, &errors.ErrInvalidTypeAssertion{
				Value: change,
				Type:  reflect.TypeOf(&nucleus.Change{}),
			}
		}

		changes[i] = &ppb.Change{
			Id:    ch.String(),
			Age:   change.Age().Microseconds(),
			State: change.State(),
		}
	}
	return changes, nil
}

func (p pndServer) Set(ctx context.Context, request *ppb.SetRequest) (*ppb.SetResponse, error) {
	pid, err := uuid.Parse(request.Pid)
	if err != nil {
		return nil, err
	}

	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, err
	}

	ondResp, err := handleSetOnd(pnd, request.Ond)
	if err != nil {
		return nil, err
	}
	changeResp, err := handleSetChange(pnd, request.Change)
	if err != nil {
		return nil, err
	}
	changeRequestResp, err := handleChangeRequest(pnd, request.ChangeRequest)
	if err != nil {
		return nil, err
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
		Responses: []*ppb.SetResponse{
			ondResp,
			changeResp,
			changeRequestResp,
		},
	}, nil
}

func handleSetOnd(pnd networkdomain.NetworkDomain, req []*ppb.SetOnd) (*ppb.SetResponse, error) {
	for _, r := range req {
		sid, err := uuid.Parse(r.Sbi.Id)
		if err != nil {
			return nil, err
		}
		if err := pnd.AddDevice(r.DeviceName, r.TransportOption, sid); err != nil {
			return nil, err
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}

func handleSetSbi(pnd networkdomain.NetworkDomain, req []*ppb.SetSbi) (*ppb.SetResponse, error) {
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_ERROR,
	}, nil
}

func handleSetChange(pnd networkdomain.NetworkDomain, req []*ppb.SetChange) (*ppb.SetResponse, error) {
	for _, r := range req {
		cuid, err := uuid.Parse(r.Cuid)
		if err != nil {
			return nil, err
		}
		switch r.Op {
		case ppb.SetChange_COMMIT:
			if err := pnd.Commit(cuid); err != nil {
				return nil, err
			}
		case ppb.SetChange_CONFIRM:
			if err := pnd.Confirm(cuid); err != nil {
				return nil, err
			}
		default:
			return nil, &errors.ErrInvalidParameters{
				Func:  handleSetChange,
				Param: r.Op,
			}
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}

func handleChangeRequest(pnd networkdomain.NetworkDomain, req []*ppb.ChangeRequest) (*ppb.SetResponse, error) {
	for _, r := range req {
		did, err := uuid.Parse(r.Id)
		if err != nil {
			return nil, err
		}
		if err := pnd.ChangeOND(did, r.ApiOp, r.Path, r.Value); err != nil {
			return nil, err
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}
