package nucleus

import (
	"context"
	"reflect"

	"code.fbi.h-da.de/cocsn/gosdn/interfaces/southbound"

	"google.golang.org/grpc"

	ppb "code.fbi.h-da.de/cocsn/api/go/gosdn/pnd"

	"code.fbi.h-da.de/cocsn/gosdn/forks/goarista/gnmi"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus/errors"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus/types"
	pathutils "code.fbi.h-da.de/cocsn/gosdn/nucleus/util/path"
	gpb "github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/gnmi/proto/gnmi_ext"
	"github.com/openconfig/goyang/pkg/yang"
	"github.com/openconfig/ygot/ygot"
	"github.com/openconfig/ygot/ytypes"
	log "github.com/sirupsen/logrus"

	tpb "code.fbi.h-da.de/cocsn/api/go/gosdn/transport"
)

var opmap = map[ppb.ApiOperation]string{
	ppb.ApiOperation_UPDATE:  "update",
	ppb.ApiOperation_REPLACE: "replace",
	ppb.ApiOperation_DELETE:  "delete",
}

// Gnmi implements the Transport interface and provides an SBI with the
// possibility to access a gNMI endpoint.
type Gnmi struct {
	SetNode   func(schema *yang.Entry, root interface{}, path *gpb.Path, val interface{}, opts ...ytypes.SetNodeOpt) error
	RespChan  chan *gpb.SubscribeResponse
	Unmarshal func([]byte, []string, interface{}, ...ytypes.UnmarshalOpt) error
	Options   *tpb.TransportOption
	client    gpb.GNMIClient
	config    *gnmi.Config
}

// newGnmiTransport takes a struct of GnmiTransportOptions and returns a Gnmi
// transport based on the values of it.
// Do not call directly. Use NewTransport() instead.
func newGnmiTransport(opts *tpb.TransportOption, sbi southbound.SouthboundInterface) (*Gnmi, error) {
	if opts == nil || sbi == nil {
		return nil, &errors.ErrInvalidParameters{}
	} else if opts.TransportOption == nil {
		return nil, &errors.ErrInvalidParameters{}
	}
	gnmiConfig := &gnmi.Config{
		Addr:        opts.Address,
		Password:    opts.Password,
		Username:    opts.Username,
		TLS:         opts.Tls,
		Compression: opts.GetGnmiTransportOption().GetCompression(),
	}
	c, err := gnmi.Dial(gnmiConfig)
	if err != nil {
		return nil, err
	}
	log.WithFields(log.Fields{
		"target": opts.Address,
		"tls":    opts.Tls,
	}).Info("building new gNMI transport")
	return &Gnmi{
		SetNode:  sbi.SetNode(),
		RespChan: make(chan *gpb.SubscribeResponse),
		Options:  opts,
		client:   c,
		config:   gnmiConfig,
	}, nil
}

func parseDialOptions(opts map[string]string) []grpc.DialOption {
	return nil
}

// Get takes a slice of gnmi paths, splits them and calls get for each one of them.
func (g *Gnmi) Get(ctx context.Context, params ...string) (interface{}, error) {
	if g.client == nil {
		return nil, &errors.ErrNilClient{}
	}
	paths := gnmi.SplitPaths(params)
	return g.get(ctx, paths, "")
}

// Set takes a slice of params. This slice must contain at least one operation.
// It can contain an additional arbitrary amount of operations and extensions.
func (g *Gnmi) Set(ctx context.Context, args ...interface{}) error {
	if g.client == nil {
		return &errors.ErrNilClient{}
	}
	if len(args) == 0 {
		return &errors.ErrInvalidParameters{
			Func:  "nucleus.Set()",
			Param: "no parameters provided",
		}
	}

	if len(args) == 2 {
		switch args[0].(type) {
		case ygot.GoStruct:
			return g.applyDiff(ctx, args[0])
		default:

		}
	}

	opts := make([]interface{}, 0)
	for _, o := range args {
		attrs, ok := o.([]string)
		if !ok {
			return &errors.ErrInvalidTypeAssertion{
				Value: o,
				Type:  reflect.TypeOf("placeholder"),
			}
		} else if attrs == nil || len(attrs) == 0 {
			return &errors.ErrInvalidParameters{
				Func:  "nucleus.Set()",
				Param: "no parameters provided",
			}
		}
		opts = append(opts, &gnmi.Operation{
			// Hardcoded TransportUpdate until multiple operations are supported
			Type:   opmap[ppb.ApiOperation_UPDATE],
			Origin: "",
			Target: "",
			Path:   gnmi.SplitPath(attrs[0]),
			Val:    attrs[1],
		})
	}

	// Loop over args and create ops and exts
	// Invalid args cause unhealable error
	ops := make([]*gnmi.Operation, 0)
	exts := make([]*gnmi_ext.Extension, 0)
	for _, p := range opts {
		switch p.(type) {
		case *gnmi.Operation:
			op := p.(*gnmi.Operation)
			if op.Target == "" {
				op.Target = g.Options.Address
			}
			ops = append(ops, op)
		case *gnmi_ext.Extension:
			exts = append(exts, p.(*gnmi_ext.Extension))
		default:
			return &errors.ErrInvalidParameters{
				Func:  "nucleus.Set()",
				Param: "args contain invalid type",
			}
		}
	}
	if len(ops) == 0 {
		return &errors.ErrInvalidParameters{
			Func:  "nucleus.Set()",
			Param: "no operations provided",
		}
	}
	resp, err := g.set(ctx, ops, exts...)
	if err != nil {
		return err
	}
	log.Info(resp)
	return nil
}

func (g *Gnmi) applyDiff(ctx context.Context, payload ...interface{}) error {
	if len(payload) != 2 {
		return &errors.ErrInvalidParameters{}
	}
	op := ctx.Value(types.CtxKeyOperation)
	oldstate, ok := payload[0].(ygot.GoStruct)
	if !ok {
		return &errors.ErrInvalidTypeAssertion{
			Value: payload[0],
			Type:  reflect.TypeOf("ygot.GoStruct"),
		}
	}
	newstate, ok := payload[1].(ygot.GoStruct)
	if !ok {
		return &errors.ErrInvalidTypeAssertion{
			Value: payload[1],
			Type:  reflect.TypeOf("ygot.GoStruct"),
		}
	}

	diff, err := ygot.Diff(oldstate, newstate)
	if err != nil {
		return err
	}
	req := &gpb.SetRequest{}
	if diff.Update != nil {
		switch op {
		case ppb.ApiOperation_UPDATE:
			req.Update = diff.Update
		case ppb.ApiOperation_REPLACE:
			req.Replace = diff.Update
		default:
			return &errors.ErrOperationNotSupported{}
		}
	} else if diff.Delete != nil {
		req.Delete = diff.Delete
	}
	resp, err := g.client.Set(ctx, req)
	log.Info(resp)
	return err
}

//Subscribe subscribes to a gNMI target
func (g *Gnmi) Subscribe(ctx context.Context, params ...string) error {
	if g.client == nil {
		return &errors.ErrNilClient{}
	}
	return g.subscribe(ctx)
}

// Type returns the gNMI transport type
func (g *Gnmi) Type() string {
	return "gnmi"
}

// ProcessResponse takes a gNMI response and serializes the contents to the root struct.
func (g *Gnmi) ProcessResponse(resp interface{}, root interface{}, s *ytypes.Schema) error {
	models := s.SchemaTree
	r := resp.(*gpb.GetResponse)
	rn := r.Notification
	for _, msg := range rn {
		for _, update := range msg.Update {
			path := update.Path
			fullPath := path
			val, ok := update.Val.Value.(*gpb.TypedValue_JsonIetfVal)
			if ok {
				opts := []ytypes.UnmarshalOpt{&ytypes.IgnoreExtraFields{}}
				return g.Unmarshal(val.JsonIetfVal, pathutils.ToStrings(fullPath), root, opts...)
			}
			// TODO(mk): Evaluate hardcoded model key
			schema := models["Device"]
			opts := []ytypes.SetNodeOpt{&ytypes.InitMissingElements{}, &ytypes.TolerateJSONInconsistencies{}}
			if err := g.SetNode(schema, root, update.Path, update.Val, opts...); err != nil {
				return err
			}
		}
	}
	return nil
}

// Capabilities calls GNMI capabilities
func (g *Gnmi) Capabilities(ctx context.Context) (interface{}, error) {
	log.WithFields(log.Fields{
		"target": g.Options.Address,
	}).Info("sending gNMI capabilities request")
	ctx = gnmi.NewContext(ctx, g.config)
	ctx = context.WithValue(ctx, types.CtxKeyConfig, g.config) //nolint
	resp, err := g.client.Capabilities(ctx, &gpb.CapabilityRequest{})
	if err != nil {
		return nil, err
	}
	return resp, nil
}

// get calls GNMI get
func (g *Gnmi) get(ctx context.Context, paths [][]string, origin string) (interface{}, error) {

	ctx = gnmi.NewContext(ctx, g.config)
	ctx = context.WithValue(ctx, types.CtxKeyConfig, g.config) //nolint
	req, err := gnmi.NewGetRequest(ctx, paths, origin)
	if err != nil {
		return nil, err
	}
	return g.getWithRequest(ctx, req)
}

// getWithRequest takes a fully formed GetRequest, performs the Get,
// and returns any response.
func (g *Gnmi) getWithRequest(ctx context.Context, req *gpb.GetRequest) (interface{}, error) {
	if req == nil {
		return nil, &errors.ErrNil{}
	}
	log.WithFields(log.Fields{
		"target": g.Options.Address,
		"path":   req.Path,
	}).Info("sending gNMI get request")

	resp, err := g.client.Get(ctx, req)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

// Set calls GNMI set
func (g *Gnmi) set(ctx context.Context, setOps []*gnmi.Operation,
	exts ...*gnmi_ext.Extension) (*gpb.SetResponse, error) {
	ctx = gnmi.NewContext(ctx, g.config)
	targets := make([]string, len(setOps))
	paths := make([][]string, len(setOps))
	values := make([]string, len(setOps))
	for i, v := range setOps {
		targets[i] = v.Target
		paths[i] = v.Path
		values[i] = v.Val
	}
	log.WithFields(log.Fields{
		"targets": targets,
		"paths":   paths,
		"values":  values,
	}).Info("sending gNMI set request")
	return gnmi.Set(ctx, g.client, setOps, exts...)
}

// Subscribe calls GNMI subscribe
func (g *Gnmi) subscribe(ctx context.Context) error {
	ctx = gnmi.NewContext(ctx, g.config)
	opts, ok := ctx.Value(types.CtxKeyOpts).(*gnmi.SubscribeOptions)
	if !ok {
		return &errors.ErrInvalidTypeAssertion{
			Value: reflect.TypeOf(ctx.Value(types.CtxKeyOpts)),
			Type:  reflect.TypeOf(&gnmi.SubscribeOptions{}),
		}
	}
	go func() {
		log.WithFields(log.Fields{
			"address":  opts.Target,
			"paths":    opts.Paths,
			"mode":     opts.Mode,
			"interval": opts.SampleInterval,
		}).Info("subscribed to gNMI target")
		for {
			resp := <-g.RespChan
			if resp != nil {
				if err := gnmi.LogSubscribeResponse(resp); err != nil {
					log.Fatal(err)
				}
			}
		}
	}()
	return gnmi.SubscribeErr(ctx, g.client, opts, g.RespChan)
}

// Close calls GNMI close
func (g *Gnmi) Close() error {
	return nil
}

// SetPassthrough allows to pass an existing SetRequest. Used for cSBI
func (g *Gnmi) SetPassthrough(ctx context.Context, req *gpb.SetRequest) (*gpb.SetResponse, error) {
	return g.client.Set(ctx, req)
}

// GetPassthrough allows to pass an existing GetRequest. Used for cSBI
func (g *Gnmi) GetPassthrough(ctx context.Context, req *gpb.GetRequest) (*gpb.GetResponse, error) {
	return g.client.Get(ctx, req)
}
