package nucleus

import (
	"reflect"

	spb "code.fbi.h-da.de/cocsn/api/go/gosdn/southbound"
	"code.fbi.h-da.de/cocsn/gosdn/interfaces/southbound"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus/errors"

	"code.fbi.h-da.de/cocsn/yang-models/generated/openconfig"
	"github.com/google/uuid"
	gpb "github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/goyang/pkg/yang"
	"github.com/openconfig/ygot/util"
	"github.com/openconfig/ygot/ygot"
	"github.com/openconfig/ygot/ytypes"
	log "github.com/sirupsen/logrus"
)

// NewSBI creates a SouthboundInterface of a given type.
func NewSBI(southbound spb.Type) southbound.SouthboundInterface {
	switch southbound {
	case spb.Type_OPENCONFIG:
		return &OpenConfig{id: uuid.New()}
	default:
		return nil
	}
}

// OpenConfig is the implementation of an OpenConfig SBI.
// The struct holds the YANG schema and a function that
// returns an SBI specific SetNode function.
type OpenConfig struct {
	schema *ytypes.Schema
	id     uuid.UUID
}

// SbiIdentifier returns the string representation of
// the SBI
// deprecated
func (oc *OpenConfig) SbiIdentifier() string {
	return "openconfig"
}

// Schema returns a ygot generated openconfig Schema as ytypes.Schema
func (oc *OpenConfig) Schema() *ytypes.Schema {
	schema, err := openconfig.Schema()
	oc.schema = schema
	if err != nil {
		log.Fatal(err)
	}
	return schema
}

// SetNode injects OpenConfig specific model
// representation to the transport.
// Needed for type assertion.
func (oc *OpenConfig) SetNode() func(schema *yang.Entry, root interface{}, path *gpb.Path, val interface{}, opts ...ytypes.SetNodeOpt) error {
	return func(schema *yang.Entry, root interface{}, path *gpb.Path, val interface{}, opts ...ytypes.SetNodeOpt) error {
		return ytypes.SetNode(schema, root.(*openconfig.Device), path, val, opts...)
	}
}

// Unmarshal injects OpenConfig specific model
// representation to the transport.
// Needed for type assertion.
func (oc *OpenConfig) Unmarshal() func([]byte, []string, interface{}, ...ytypes.UnmarshalOpt) error {
	return unmarshal
}

// unmarshal parses a root or 1st level gNMI response to a go struct
// Named return to return appropriate recover error
func unmarshal(bytes []byte, fields []string, goStruct interface{}, opt ...ytypes.UnmarshalOpt) (err error) {
	defer func() {
		if r := recover(); r != nil {
			err = r.(error)
		}
	}()
	switch l := len(fields); l {
	case 0:
		return openconfig.Unmarshal(bytes, goStruct.(*openconfig.Device), opt...)
	case 1:
	default:
		return &errors.ErrUnsupportedPath{Path: fields}
	}
	var c ygot.GoStruct
	var field string

	// Load SBI definition
	d := openconfig.Device{}
	c, field, err = iter(&d, fields)
	if err != nil {
		return
	}
	if err = openconfig.Unmarshal(bytes, c, opt...); err != nil {
		return
	}
	reflect.ValueOf(goStruct.(*openconfig.Device)).Elem().FieldByName(field).Set(reflect.ValueOf(c))
	return nil
}

// iter walks down the provided paths and initializes the ygot.GoStruct. It only works for
// the root level. Named returns to return appropriate recover error
// TODO(mk): Fix deeper layers
func iter(a ygot.GoStruct, fields []string) (b ygot.GoStruct, f string, err error) {
	defer func() {
		if r := recover(); r != nil {
			err = r.(error)
		}
	}()
	var c ygot.GoStruct
	var configStruct reflect.Value
	f = fields[0]
	s := reflect.ValueOf(a).Elem()
	h := s.FieldByName(f)
	configStruct = reflect.New(h.Type())

	// Pointer of field needs to be initialized.
	// Very convoluted russian doll trick
	// https://stackoverflow.com/a/57469950/4378176
	// https://golang.org/src/encoding/json/decode.go?s#L474
	// TODO(mk): Prettify
	p2 := configStruct.Elem()
	// If we have KeyHelperGoStruct we need make and modify map instead of plain struct
	if p2.Kind() == reflect.Map {
		p2.Set(reflect.MakeMap(p2.Type()))
		configStruct.Elem().Set(p2)
		if err := util.InsertIntoMapStructField(a, f, "", p2); err != nil {
			panic(err)
		}
	} else {
		configStruct.Elem().Set(reflect.New(p2.Type().Elem()))
		b = configStruct.Elem().Interface().(ygot.GoStruct)
	}
	if len(fields) > 1 {
		c, _, _ = iter(b, fields[1:])
	} else {
		return
	}
	reflect.ValueOf(b).Elem().FieldByName(f).Set(reflect.ValueOf(c))
	return
}

// ID returns the ID of the OpenConfig SBI
func (oc *OpenConfig) ID() uuid.UUID {
	return oc.id
}

// Type returns the Southbound's type
func (oc *OpenConfig) Type() spb.Type { return spb.Type_OPENCONFIG }
