package nucleus

import (
	"fmt"
	"reflect"
	"sync"

	"code.fbi.h-da.de/cocsn/gosdn/interfaces/device"
	"code.fbi.h-da.de/cocsn/gosdn/interfaces/networkdomain"
	"code.fbi.h-da.de/cocsn/gosdn/interfaces/southbound"
	"code.fbi.h-da.de/cocsn/gosdn/interfaces/store"
	"code.fbi.h-da.de/cocsn/gosdn/nucleus/errors"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

var storeLock sync.RWMutex

// NewStore returns a generic Store
func NewStore() store.Store {
	return genericStore{}
}

// NewPndStore returns a PndStore
func NewPndStore() *PndStore {
	return &PndStore{genericStore{}}
}

// NewSbiStore returns a SbiStore
func NewSbiStore() *SbiStore {
	return &SbiStore{genericStore{}}
}

// NewDeviceStore returns a DeviceStore
func NewDeviceStore() *DeviceStore {
	return &DeviceStore{genericStore: genericStore{}, deviceNameToUUIDLookup: make(map[string]uuid.UUID)}
}

// NewChangeStore returns a ChangeStore
func NewChangeStore() *ChangeStore {
	return &ChangeStore{genericStore{}}
}

type genericStore map[uuid.UUID]store.Storable

// Exists takes a Storable's UUID and checks its existence in the store.
func (s genericStore) Exists(id uuid.UUID) bool {
	storeLock.RLock()
	defer storeLock.RUnlock()
	_, ok := s[id]
	return ok
}

// Add adds a Storable to the Store
func (s genericStore) Add(item store.Storable) error {
	if s.Exists(item.ID()) {
		return &errors.ErrAlreadyExists{Item: item}
	}
	storeLock.Lock()
	s[item.ID()] = item
	storeLock.Unlock()
	log.WithFields(log.Fields{
		"type": reflect.TypeOf(item),
		"uuid": item.ID(),
	}).Debug("storable was added")
	return nil
}

// Get takes a Storable's UUID and returns the Storable. If the requested
// Storable does not exist an error is returned. Get is only type safe for
// this Storable interface. For type safe get operations on specialised stores
// use GetDevice, GetPND, GetSBI, or GetChange respectively.
func (s genericStore) Get(id uuid.UUID) (store.Storable, error) {
	if !s.Exists(id) {
		return nil, &errors.ErrNotFound{ID: id}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("storable was accessed")
	storeLock.RLock()
	defer storeLock.RUnlock()
	return s[id], nil
}

// Delete takes a Storable's UUID and deletes it. If the specified UUID does not
// exist in the Store an error is returned.
func (s genericStore) Delete(id uuid.UUID) error {
	if !s.Exists(id) {
		return &errors.ErrNotFound{ID: id}
	}
	storeLock.Lock()
	delete(s, id)
	storeLock.Unlock()
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("storable was deleted")
	return nil
}

// UUIDs returns all UUIDs in the store.
func (s genericStore) UUIDs() []uuid.UUID {
	storeLock.RLock()
	defer storeLock.RUnlock()
	keys := make([]uuid.UUID, len(s))
	i := 0
	for k := range s {
		keys[i] = k
		i++
	}
	return keys
}

// SbiStore is used to store SouthboundInterfaces
type SbiStore struct {
	genericStore
}

// GetSBI takes a SouthboundInterface's UUID and returns the SouthboundInterface. If the requested
// SouthboundInterface does not exist an error is returned.
func (s SbiStore) GetSBI(id uuid.UUID) (southbound.SouthboundInterface, error) {
	item, err := s.genericStore.Get(id)
	if err != nil {
		return nil, err
	}
	sbi, ok := item.(southbound.SouthboundInterface)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: sbi,
			Type:  "SouthboundInterface",
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("southbound interface was accessed")
	return sbi, nil
}

// PndStore is used to store PrincipalNetworkDomains
type PndStore struct {
	genericStore
}

// GetPND takes a PrincipalNetworkDomain's UUID and returns the PrincipalNetworkDomain. If the requested
// PrincipalNetworkDomain does not exist an error is returned.
func (s PndStore) GetPND(id uuid.UUID) (networkdomain.NetworkDomain, error) {
	item, err := s.genericStore.Get(id)
	if err != nil {
		return nil, err
	}
	pnd, ok := item.(networkdomain.NetworkDomain)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: pnd,
			Type:  "PrincipalNetworkDomain",
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("principal network domain was accessed")
	return pnd, nil
}

// DeviceStore is used to store Devices
type DeviceStore struct {
	deviceNameToUUIDLookup map[string]uuid.UUID
	genericStore
}

// GetDevice takes a Device's UUID and returns the Device. If the requested
// Device does not exist an error is returned.
func (s DeviceStore) GetDevice(id uuid.UUID, parseErrors ...error) (device.Device, error) {
	var foundID uuid.UUID

	foundID = id

	for _, parseErrs := range parseErrors {
		if parseErrs != nil {
			switch e := parseErrs.(type) {
			case *errors.ErrInvalidUUID:
				myID, ok := s.deviceNameToUUIDLookup[e.DeviceName]
				if !ok {
					log.Debug(fmt.Sprintf("no device named %s found", foundID))
					return nil, &errors.ErrNotFound{}
				}

				foundID = myID
			}
		}
	}

	item, err := s.genericStore.Get(foundID)
	if err != nil {
		return nil, err
	}
	d, ok := item.(device.Device)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: d,
			Type:  reflect.TypeOf((device.Device)(nil)),
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
		"name": d.Name,
	}).Debug("device was accessed")

	return d, nil
}

// FromString is a helper to check if a provided string as a valid UUID or a name.
func FromString(id string) (uuid.UUID, error) {
	idAsUUID, err := uuid.Parse(id)

	// id is no UUID therefore it could be a device name.
	// The name will be returned within the error.
	if err != nil {
		log.WithFields(log.Fields{
			"identifier": id,
		}).Debug(err)
		return uuid.Nil, &errors.ErrInvalidUUID{DeviceName: id}
	}

	return idAsUUID, nil
}

// Add adds a device to the device store.
// It also adds the name of the device to the lookup table.
func (s DeviceStore) Add(item store.Storable, name string) error {
	if s.Exists(item.ID()) {
		return &errors.ErrAlreadyExists{Item: item}
	}

	s.deviceNameToUUIDLookup[name] = item.ID()

	storeLock.Lock()
	s.genericStore[item.ID()] = item
	storeLock.Unlock()

	log.WithFields(log.Fields{
		"type": reflect.TypeOf(item),
		"uuid": item.ID(),
	}).Debug("storable was added")

	return nil
}

// Delete deletes a device from the device store.
func (s DeviceStore) Delete(id uuid.UUID) error {
	if !s.Exists(id) {
		return &errors.ErrNotFound{ID: id}
	}

	storeLock.Lock()
	delete(s.genericStore, id)
	storeLock.Unlock()

	for key, value := range s.deviceNameToUUIDLookup {
		if value == id {
			delete(s.deviceNameToUUIDLookup, key)
		}
	}

	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("storable was deleted")

	return nil
}

// ChangeStore is used to store Changes
type ChangeStore struct {
	genericStore
}

// GetChange takes a Change's UUID and returns the Change. If the requested
// Change does not exist an error is returned.
func (s ChangeStore) GetChange(id uuid.UUID) (*Change, error) {
	item, err := s.genericStore.Get(id)
	if err != nil {
		return nil, err
	}
	change, ok := item.(*Change)
	if !ok {
		return nil, &errors.ErrInvalidTypeAssertion{
			Value: change,
			Type:  reflect.TypeOf(&Change{}),
		}
	}
	log.WithFields(log.Fields{
		"uuid": id,
	}).Debug("change was accessed")
	return change, nil
}
